use std::time::{SystemTime, UNIX_EPOCH};
use sha1::Sha1;
use hmac::{Hmac, Mac};
use base32::Alphabet;

type HmacSha1 = Hmac<Sha1>;

pub trait TOTP{
    fn get_totp(&mut self) -> u32;
    fn set_new_totp(&mut self, now: &u64);
}

pub struct TOTPData{
    pub time_validity: u64,
    pub seed: String,
    pub last_totp: u32,
    pub generation_time: u64,
    pub issuer: String,
    pub digits: u32
}

impl TOTP for TOTPData {
    fn get_totp(&mut self) -> u32 {
        let now:u64 = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .expect("")
            .as_secs();

        if self.generation_time + self.time_validity < now {
            self.set_new_totp(&now);
        }
        return self.last_totp.clone();
    }

    fn set_new_totp(&mut self, now: &u64) {
        let t = (now/self.time_validity).to_be_bytes();
        self.generation_time = now.clone();
        let key = base32::decode(Alphabet::Rfc4648 {padding: false}, self.seed.as_str()).expect("Invalid Base32 encoding");
        let mut mac = HmacSha1::new_from_slice(&key).expect("HMAC can take key of any size");
        mac.update(&t);
        let result = mac.finalize().into_bytes();
        let offset = (result[result.len() -1] & 0x0F) as usize;
        let truncated_hash = &result[offset..offset+4];
        let binary_code = ((truncated_hash[0] as u32) << 24)
            | ((truncated_hash[1] as u32) << 16)
            | ((truncated_hash[2] as u32) << 8)
            | (truncated_hash[3] as u32);
        let otp = (binary_code & 0x7FFFFFFF) % 10_u32.pow(self.digits); 
        self.last_totp = otp
    }
}

